import pandas as pd

import bionty as bt


def test_efo_readout_inspect_ontology_id():
    df = pd.DataFrame(
        index=[
            "EFO:0000001",
            "EFO:1002050",
            "EFO:1002047",
            "EFO:0000005",
            "This readout does not exist",
        ]
    )

    ro = bt.Readout(source="efo", version="3.48.0")
    inspected_df = ro.inspect(df.index, ro.ontology_id, return_df=True)

    inspect = inspected_df["__mapped__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)


def test_readout_parse():
    ro = bt.Readout(source="efo", version="3.48.0")

    assert ro._parse("EFO:0008913") == {
        "ontology_id": "EFO:0008913",
        "name": "single-cell RNA sequencing",
        "molecule": "RNA assay",
        "instrument": "single cell sequencing",
        "measurement": None,
    }



from typing import Literal, Optional, Tuple, Union

import pandas as pd

from .._bionty import Bionty, BiontyField
from ._shared_docstrings import _doc_params, doc_entites


@_doc_params(doc_entities=doc_entites)
class Gene(Bionty):
    """Gene.

    1. Ensembl
    Edits of terms are coordinated and reviewed on:
    https://www.ensembl.org/

    Args:
        {doc_entities}

    Notes:
        Biotypes: https://www.ensembl.org/info/genome/genebuild/biotypes.html
        Gene Naming: https://www.ensembl.org/info/genome/genebuild/gene_names.html
    """

    def __init__(
        self,
        species: str = "human",
        source: Optional[Literal["ensembl"]] = None,
        version: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            source=source,
            version=version,
            species=species,
            **kwargs,
        )

    def lookup(self, field: Union[BiontyField, str] = "symbol") -> Tuple:
        """Return an auto-complete object for the bionty field.

        Args:
            field: The field to lookup the values for.
                   Defaults to 'name'.

        Returns:
            A NamedTuple of lookup information of the field values.

        Examples:
            >>> import bionty as bt
            >>> gene_lookout = bt.Gene().lookup()
            >>> gene_lookout.TEF
        """
        return super().lookup(field=field)

    def search(
        self,
        string: str,
        field: Union[BiontyField, str] = "symbol",
        synonyms_field: Union[BiontyField, str, None] = "synonyms",
        case_sensitive: bool = True,
        return_ranked_results: bool = False,
    ) -> pd.DataFrame:
        """Fuzzy matching of a given string against a Bionty field.

        Args:
            string: The input string to match against the field ontology values.
            field: The BiontyField of ontology the input string is matching against.
            synonyms_field: Also map against in the synonyms (If None, no mapping against synonyms).
            case_sensitive: Whether the match is case sensitive.
            return_ranked_results: Whether to return all entries ranked by matching ratios.

        Returns:
            Best match of the input string.

        Examples:
            >>> import bionty as bt
            >>> celltype_bionty = bt.CellType()
            >>> celltype_bionty.search("gamma delta T cell", celltype_bionty.name)
        """
        return super().search(
            string=string,
            field=field,
            synonyms_field=synonyms_field,
            case_sensitive=case_sensitive,
            return_ranked_results=return_ranked_results,
        )



from typing import Literal, Optional

import pandas as pd

from bionty.entities._shared_docstrings import _doc_params, species_removed

from .._bionty import Bionty


@_doc_params(doc_entities=species_removed)
class Species(Bionty):
    """Species.

    1. Species ontology
    Edits of terms are coordinated and reviewed on:
    https://www.ensembl.org/index.html

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        source: Optional[Literal["ensembl"]] = None,
        version: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(source=source, version=version, **kwargs)

    def _load_df(self) -> pd.DataFrame:
        if not self._local_parquet_path.exists():
            self._url_download(self._url, self._local_ontology_path)
            df = pd.read_csv(self._local_ontology_path, sep="\t", index_col=False)
            df.rename(
                columns={
                    "#name": "name",
                    "species": "scientific_name",
                    "taxonomy_id": "taxon_id",
                },
                inplace=True,
            )
            df["name"] = df["name"].str.lower()
            df.insert(0, "id", "NCBI_" + df["taxon_id"].astype(str))
            df.to_parquet(self._local_parquet_path)
            return df
        else:
            return pd.read_parquet(self._local_parquet_path)

    def df(self) -> pd.DataFrame:
        """Pandas DataFrame of the ontology.

        Returns:
            A Pandas DataFrame of the ontology.

        Examples:
            >>> import bionty as bt
            >>> bt.Species().df()
        """
        return self._df.set_index("name")



from filelock import FileLock  # type: ignore

from .dev._handle_sources import (
    ROOT,
    create_currently_used_sources_yaml,
    create_or_update_sources_local_yaml,
)

# Make this code safe when running bionty from multiple processes
with FileLock(ROOT / "bionty.lock"):
    create_or_update_sources_local_yaml(overwrite=False)
    # always generate a new CURRENT_SOURCES file
    create_currently_used_sources_yaml(overwrite=True)



from typing import Literal, Optional

from .._bionty import Bionty
from ._shared_docstrings import _doc_params, doc_entites


@_doc_params(doc_entities=doc_entites)
class CellLine(Bionty):
    """Cell line.

    1. Cell Line Ontology
    Edits of terms are coordinated and reviewed on:
    https://github.com/CLO-ontology/CLO

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        species: Optional[str] = None,
        source: Optional[Literal["clo"]] = None,
        version: Optional[str] = None,
        **kwargs
    ) -> None:
        super().__init__(
            source=source,
            version=version,
            species=species,
            include_id_prefixes={"clo": ["CLO"]},
            **kwargs
        )



from pathlib import Path

from bionty.dev._io import url_download


def test_url_download():
    url = "https://bionty-assets.s3.amazonaws.com/bfxpipelines.json"
    localpath = Path(f"./{url.split('/')[-1]}")
    assert not localpath.exists()

    localpath_default = url_download(url=url)
    assert Path(localpath_default) == localpath
    assert localpath.exists()



from pathlib import Path
from typing import Dict, List, Literal, Union

import pandas as pd
from lamin_logger import logger

from bionty._settings import settings
from bionty.dev._io import load_yaml, write_yaml

ROOT = Path(__file__).parent.parent / "sources"
PUBLIC_SOURCES = ROOT / "sources.yaml"

# hidden from the users
CURRENT_SOURCES = ROOT / ".current_sources.yaml"
LAMINDB_SOURCES = ROOT / ".lamindb_current_sources.yaml"

# Visible to the users and can be modified
LOCAL_SOURCES = settings.versionsdir / "sources_local.yaml"


def LAMINDB_INSTANCE_LOADED():
    return (Path.home() / ".lamin/current_instance.env").exists()


def reset_sources(confirm: bool = False):
    """Reset local bionty sources file."""
    if confirm:
        from importlib import reload

        import bionty

        try:
            LOCAL_SOURCES.unlink()
            logger.success(f"Removed file: {LOCAL_SOURCES}.")
        except FileNotFoundError:
            pass
        try:
            CURRENT_SOURCES.unlink()
            logger.success(f"Removed file: {CURRENT_SOURCES}.")
        except FileNotFoundError:
            pass
        try:
            LAMINDB_SOURCES.unlink()
            logger.success(f"Removed file: {LAMINDB_SOURCES}.")
        except FileNotFoundError:
            pass

        reload(bionty)
        logger.info("Reloaded bionty")

    else:
        logger.warning(
            "Are you sure to reset your local bionty sources? Pass 'confirm=True'"
        )


def create_or_update_sources_local_yaml(overwrite: bool = True) -> None:
    """If LOCAL_SOURCES doesn't exist, copy from PUBLIC_SOURCES and create it.

    Args:
        overwrite: Whether to overwrite the current LOCAL_SOURCES.
    """
    if not LOCAL_SOURCES.exists() or overwrite:
        public_df_records = parse_sources_yaml(PUBLIC_SOURCES).to_dict(  # type: ignore
            orient="records"
        )
        versions = add_records_to_existing_dict(public_df_records, {})
        versions_header = {"version": load_yaml(PUBLIC_SOURCES).get("version")}
        versions_header.update(versions)
        write_yaml(versions_header, LOCAL_SOURCES)
    else:
        update_local_from_public_sources_yaml()


def parse_sources_yaml(filepath: Union[str, Path]) -> pd.DataFrame:
    """Parse values from sources yaml file into a DataFrame.

    Args:
        filepath: Path to the versions yaml file.
        return_df: Whether to return a Pandas DataFrame

    Returns:
        - entity
        - source
        - species
        - version
        - url
        - md5
        - source_name
        - source_website
    """
    all_rows = []
    for entity, sources in load_yaml(filepath).items():
        if entity == "version":
            continue
        for source, species_source in sources.items():
            name = species_source.get("name", "")
            website = species_source.get("website", "")
            for species, versions in species_source.items():
                if species in ["name", "website"]:
                    continue
                for version_key, version_meta in versions.items():
                    row = (
                        entity,
                        source,
                        species,
                        str(version_key),
                        version_meta.get("url"),
                        version_meta.get("md5", ""),
                        name,
                        website,
                    )
                    all_rows.append(row)

    return pd.DataFrame(
        all_rows,
        columns=[
            "entity",
            "source",
            "species",
            "version",
            "url",
            "md5",
            "source_name",
            "source_website",
        ],
    )


def create_currently_used_sources_yaml(
    overwrite: bool = True, source: Literal["versions", "local"] = "local"
) -> None:
    """Write the most recent version to the CURRENT_SOURCES .

    Takes the 1st source defined in the source.

    Args:
        overwrite: Whether to overwrite the CURRENT_SOURCES even if it exists already.
        source: The yaml source to use to create the CURRENT_SOURCES.
                Defaults to 'local'.
    """
    if not CURRENT_SOURCES.exists() or overwrite:
        source_path = PUBLIC_SOURCES if source == "versions" else LOCAL_SOURCES

        write_yaml(parse_currently_used_sources(source_path), CURRENT_SOURCES)


def records_diff_btw_yamls(
    yamlpath1: Union[str, Path], yamlpath2: Union[str, Path]
) -> List:
    """Records in yaml1 but not yaml2."""
    public_df_records = parse_sources_yaml(yamlpath1).to_dict(orient="records")
    local_df_records = parse_sources_yaml(yamlpath2).to_dict(orient="records")
    additional_records = [
        record for record in public_df_records if record not in local_df_records
    ]

    return additional_records


def update_local_from_public_sources_yaml() -> None:
    """Update LOCAL_SOURCES to add additional entries from PUBLIC_SOURCES."""
    additional_records = records_diff_btw_yamls(PUBLIC_SOURCES, LOCAL_SOURCES)
    if len(additional_records) > 0:
        updated_local_versions = add_records_to_existing_dict(
            additional_records, load_yaml(LOCAL_SOURCES)
        )
        write_yaml(updated_local_versions, LOCAL_SOURCES)
        logger.success(
            f"New records found in the public sources.yaml, updated {LOCAL_SOURCES}!"
        )


def parse_currently_used_sources(yaml: Union[str, Path, List[Dict]]) -> Dict:
    """Parse out the most recent versions from yaml."""
    if isinstance(yaml, (str, Path)):
        df = parse_sources_yaml(yaml)
        df_current = (
            df[["entity", "source", "species", "version"]]  # type: ignore
            .drop_duplicates(["entity", "species", "source"], keep="first")
            .groupby(["entity", "species", "source"], sort=False)
            .max()
        )
        records = df_current.reset_index().to_dict(orient="records")
    else:
        records = yaml

    current_dict: Dict = {}
    for kwargs in records:
        entity, species, source, version = (
            kwargs["entity"],
            kwargs["species"],
            kwargs["source"],
            kwargs["version"],
        )
        if entity not in current_dict:
            current_dict[entity] = {}
        if species not in current_dict[entity]:
            current_dict[entity][species] = {source: version}
    return current_dict


def add_records_to_existing_dict(records: List[Dict], target_dict: Dict) -> Dict:
    """Add records to a versions yaml file."""
    for kwargs in records:
        entity, source, species, version = (
            kwargs["entity"],
            kwargs["source"],
            kwargs["species"],
            kwargs["version"],
        )
        if entity not in target_dict:
            target_dict[entity] = {}
        if source not in target_dict[entity]:
            target_dict[entity][source] = {
                species: {version: {"url": kwargs["url"], "md5": kwargs["md5"]}}
            }
            target_dict[entity][source].update(
                {"name": kwargs["source_name"], "website": kwargs["source_website"]}
            )
        if species not in target_dict[entity][source]:
            target_dict[entity][source][species] = {
                version: {"url": kwargs["url"], "md5": kwargs["md5"]}
            }
        if version not in target_dict[entity][source][species]:
            target_dict[entity][source][species][version] = {
                "url": kwargs["url"],
                "md5": kwargs["md5"],
            }
    return target_dict



from typing import Literal, Optional

from .._bionty import Bionty
from ._shared_docstrings import _doc_params, doc_entites


@_doc_params(doc_entities=doc_entites)
class Disease(Bionty):
    """Disease ontologies.

    1. Mondo
    Edits of terms are coordinated and reviewed on:
    https://github.com/monarch-initiative/mondo

    2. Human Disease Ontology
    Edits of terms are coordinated and reviewed on:
    https://github.com/DiseaseOntology/HumanDiseaseOntology

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        species: Optional[str] = None,
        source: Optional[Literal["mondo", "doid"]] = None,
        version: Optional[str] = None,
        **kwargs
    ) -> None:
        super().__init__(
            source=source,
            version=version,
            species=species,
            include_id_prefixes={"mondo": ["MONDO"]},
            **kwargs
        )



import pandas as pd

import bionty as bt


def test_mondo_disease_inspect_name():
    df = pd.DataFrame(
        index=[
            "supraglottis cancer",
            "alexia",
            "trigonitis",
            "paranasal sinus disorder",
            "This disease does not exist",
        ]
    )

    ds = bt.Disease(source="mondo", version="2023-02-06")
    inspected_df = ds.inspect(df.index, field=ds.name, return_df=True)

    inspect = inspected_df["__mapped__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)


def test_doid_disease_inspect_ontology_id():
    df = pd.DataFrame(
        index=[
            "DOID:0001816",
            "DOID:0002116",
            "DOID:5547",
            "DOID:5551",
            "This disease does not exist",
        ]
    )

    ds = bt.Disease(source="doid", version="2023-01-30")
    inspected_df = ds.inspect(df.index, field=ds.ontology_id, return_df=True)

    inspect = inspected_df["__mapped__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)



import base64
import glob
import hashlib
import json
import os
import secrets
import string
from typing import Dict, Optional

from github import Github
from rich.progress import track

BASE_BFX_PIPELINES_PATH = "bionty/versions/bfxpipelines_data"


def base62(n_char: int) -> str:
    """Like nanoid without hyphen and underscore."""
    alphabet = string.digits + string.ascii_letters.swapcase()
    id = "".join(secrets.choice(alphabet) for i in range(n_char))
    return id


def to_b64_str(bstr: bytes) -> str:
    b64 = base64.urlsafe_b64encode(bstr).decode().strip("=")
    return b64


def hash_str(s: str) -> str:
    bstr = s.encode("utf-8")
    # as we're truncating at a short length, we choose md5 over sha512
    return to_b64_str(hashlib.md5(bstr).digest())


def hash_id(input_id: Optional[str] = None, *, n_char: int) -> str:
    if input_id is None:
        return base62(n_char=n_char)
    else:
        return hash_str(input_id)[:n_char].replace("_", "0").replace("-", "0")


def generate_nf_core_pipelines_info() -> None:
    """Generates a json file that contains all required pipelines information by querying the nf-core Github org."""
    gh_login = Github(os.getenv("GITHUB_TOKEN"))
    nf_core_org = gh_login.get_organization("nf-core")
    blacklist = ["cookiecutter", "tools"]
    nf_core_pipelines = {}

    for repo in track(
        nf_core_org.get_repos(),
        description="Fetching information from nf-core repositories...",
    ):
        if "pipeline" in [topic for topic in repo.get_topics()]:
            if repo.name in blacklist:
                continue

            for version in repo.get_releases():
                actual_version = (
                    version.tag_name if len(version.tag_name) >= 1 else "pre-release"
                )
                pipeline_name = f"{repo.name} v{actual_version}"
                underscore_pipeline_name = (
                    pipeline_name.replace(" ", "_").replace(".", "_").replace("-", "_")
                )

                nf_core_pipelines[underscore_pipeline_name] = {
                    "id": hash_id(pipeline_name, n_char=12),
                    "name": pipeline_name,
                    "versions": actual_version,
                    "reference": repo.url,
                }

    with open(f"{BASE_BFX_PIPELINES_PATH}/nf_core_pipelines_info.json", "w") as f:
        json_data = json.dumps(nf_core_pipelines, indent=4)
        f.write(json_data)


def merge_json_files(pipelines_folder_path: str, output_path: str) -> None:
    """Merge all JSON files in a folder and write the merged data to a new JSON file.

    Args:
        pipelines_folder_path: Path to the folder containing the JSON files.
        output_path: Path to the output JSON file.
    """
    file_paths = glob.glob(pipelines_folder_path + "/*.json")

    pipeline_json: Dict = {}

    for file_path in file_paths:
        with open(file_path, "r") as f:
            if not file_path.endswith("bfxpipelines.json"):
                pipelines_info = json.load(f)
                pipeline_json = {**pipeline_json, **pipelines_info}

    with open(output_path, "w") as f:
        json.dump(pipeline_json, f, indent=4)


generate_nf_core_pipelines_info()
merge_json_files(
    pipelines_folder_path=BASE_BFX_PIPELINES_PATH,
    output_path=f"{BASE_BFX_PIPELINES_PATH}/bfxpipelines.json",
)



from functools import wraps
from pathlib import Path
from typing import Union

HOME_DIR = Path(f"{Path.home()}/.lamin/bionty").resolve()
ROOT_DIR = Path(__file__).parent.resolve()


def check_datasetdir_exists(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        settings.datasetdir.mkdir(exist_ok=True)
        return f(*args, **kwargs)

    return wrapper


def check_dynamicdir_exists(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        settings.dynamicdir.mkdir(exist_ok=True)
        return f(*args, **kwargs)

    return wrapper


class Settings:
    def __init__(
        self,
        datasetdir: Union[str, Path] = ROOT_DIR / "data/",
        dynamicdir: Union[str, Path] = ROOT_DIR / "_dynamic/",
        versionsdir: Union[str, Path] = HOME_DIR / "versions/",
    ):
        # setters convert to Path and resolve:
        self.datasetdir = datasetdir
        self.dynamicdir = dynamicdir
        self.versionsdir = versionsdir

        self.versionsdir.mkdir(exist_ok=True, parents=True)  # type: ignore

    @property
    def datasetdir(self):
        """Directory for datasets."""
        return self._datasetdir

    @datasetdir.setter
    def datasetdir(self, datasetdir: Union[str, Path]):
        self._datasetdir = Path(datasetdir).resolve()

    @property
    def dynamicdir(self):
        """Directory for datasets."""
        return self._dynamicdir

    @dynamicdir.setter
    def dynamicdir(self, dynamicdir: Union[str, Path]):
        self._dynamicdir = Path(dynamicdir).resolve()

    @property
    def versionsdir(self):
        """Directory for version yamls."""
        return self._versionsdir

    @versionsdir.setter
    def versionsdir(self, versionsdir: Union[str, Path]):
        self._versionsdir = Path(versionsdir).resolve()


settings = Settings()



from typing import Literal, Optional

import pandas as pd

from .._bionty import Bionty
from ._shared_docstrings import _doc_params, doc_entites


@_doc_params(doc_entities=doc_entites)
class CellMarker(Bionty):
    """Cell markers.

    1. Cell Marker Ontology
    Edits of terms are coordinated and reviewed on:
    http://bio-bigdata.hrbmu.edu.cn/CellMarker/

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        species: Optional[str] = "human",
        source: Optional[Literal["cellmarker"]] = None,
        version: Optional[str] = None,
        **kwargs
    ) -> None:
        super().__init__(source=source, version=version, species=species, **kwargs)

    def _load_df(self) -> pd.DataFrame:
        """DataFrame.

        See ingestion: https://lamin.ai/docs/bionty-assets/ingest/cell-marker-2.0
        """
        df = super()._load_df()

        # TODO: remove after updating to new version
        df = df.drop(columns=["id", "version"], errors="ignore")

        return df

    def df(self) -> pd.DataFrame:
        """Pandas DataFrame of the ontology.

        Returns:
            A Pandas DataFrame of the ontology.

        Examples:
            >>> import bionty as bt
            >>> bt.CellMarker().df()
        """
        return self._df.set_index("name")



import pandas as pd

import bionty as bt


def test_cellmarker_cellmarker_inspect_name_human():
    df = pd.DataFrame(
        index=["CCR7", "CD69", "CD8A", "CD45RA", "This protein does not exist"]
    )

    cm = bt.CellMarker(source="cellmarker", version="2.0")
    curated = cm.inspect(df.index, field=cm.name)

    assert curated == {
        "mapped": ["CCR7", "CD69", "CD8A", "CD45RA"],
        "not_mapped": ["This protein does not exist"],
    }


def test_cellmarker_cellmarker_inspect_name_mouse():
    df = pd.DataFrame(
        index=["Tcf4", "Cd36", "Cd34", "Cd45", "This protein does not exist"]
    )

    cm = bt.CellMarker(source="cellmarker", version="2.0", species="mouse")
    inspected_df = cm.inspect(df.index, field=cm.name, return_df=True)

    inspect = inspected_df["__mapped__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)



from functools import cached_property
from typing import Literal, Optional

import pandas as pd
from lamin_logger import logger

from bionty.entities._shared_docstrings import _doc_params, species_removed

from .._bionty import Bionty
from .._ontology import Ontology
from ..dev._io import s3_bionty_assets


@_doc_params(doc_entities=species_removed)
class Readout(Bionty):
    """Experimental Factor.

    1. Experimental Factor Ontology
    Edits of terms are coordinated and reviewed on:
    https://www.ebi.ac.uk/ols/ontologies/efo

    Args:
        {doc_entities}

    Also see: `bionty.Bionty <https://lamin.ai/docs/bionty/bionty.entity>`__
    """

    def __init__(
        self,
        source: Optional[Literal["efo"]] = None,
        version: Optional[str] = None,
        **kwargs,
    ) -> None:
        self._prefix = "http://www.ebi.ac.uk/efo/"
        self._readout_terms = {
            "assay": "OBI:0000070",
            "assay_by_molecule": "EFO:0002772",
            "assay_by_instrument": "EFO:0002773",
            "assay_by_sequencer": "EFO:0003740",
            "measurement": "EFO:0001444",
        }
        super().__init__(
            source=source,
            version=version,
            include_id_prefixes={"efo": ["EFO", "http://www.ebi.ac.uk/efo/"]},
            **kwargs,
        )

    @cached_property
    def ontology(self) -> Ontology:  # type:ignore
        """The Pronto Ontology object.

        See: https://pronto.readthedocs.io/en/stable/api/pronto.Ontology.html
        """
        self._download_ontology_file(
            localpath=self._local_ontology_path,  # type:ignore
            url=self._url,
            md5=self._md5,
        )
        return Ontology(handle=self._local_ontology_path, prefix=self._prefix)

    def _load_df(self) -> pd.DataFrame:
        # Extra parsing steps for EFO ontology
        if self.source == "efo":
            # Download and sync from s3://bionty-assets
            s3_bionty_assets(
                filename=self._parquet_filename,
                assets_base_url="s3://bionty-assets",
                localpath=self._local_parquet_path,
            )
            # If download is not possible, write a parquet file from ontology
            if not self._local_parquet_path.exists():
                # write df to parquet file
                df = self.ontology.to_df(
                    source=self.source, include_id_prefixes=self.include_id_prefixes
                ).reset_index()
                # fix ontology_id before saving to parquet
                df["ontology_id"] = [
                    i.replace(self._prefix, "").replace("_", ":")
                    for i in df["ontology_id"]
                ]
                df["children"] = [
                    [j.replace(self._prefix, "").replace("_", ":") for j in i]
                    for i in df["children"]
                ]
                # parse terms
                logger.info("Parsing EFO terms for the first time will take 6-10min...")
                parsed_results = []
                for term in df["ontology_id"]:
                    parsed_results.append(self._parse(term))
                df_parsed = pd.DataFrame.from_records(parsed_results)
                df = df.merge(df_parsed).set_index("ontology_id")

                df.to_parquet(self._local_parquet_path)

            # loads the df and set index
            return pd.read_parquet(self._local_parquet_path)
        else:
            return super()._load_df()

    def _parse(self, term_id: str) -> dict:
        """Parse readout attributes from EFO."""

        def _list_to_str(lst: list):
            if len(lst) == 0:
                return None
            elif len(lst) == 1:
                return lst[0].name  # type: ignore
            else:
                return ";".join([i.name for i in lst])

        def _list_subclasses(ontology: Ontology, term, *, distance=1, with_self=False):
            """Subclasses of a term."""
            termclass = ontology.get_term(term)
            return list(termclass.subclasses(distance=distance, with_self=with_self))

        term = self.ontology.get_term(term_id)
        superclasses = term.superclasses()

        # assay = self.ontology._list_subclasses(self._readout_terms["assay"])
        assay_by_molecule = _list_subclasses(
            self.ontology, self._readout_terms["assay_by_molecule"]
        )
        assay_by_instrument = _list_subclasses(
            self.ontology, self._readout_terms["assay_by_instrument"]
        )

        assay_by_sequencer = _list_subclasses(
            self.ontology, self._readout_terms["assay_by_sequencer"]
        )
        measurement = _list_subclasses(
            self.ontology, self._readout_terms["measurement"]
        )

        # get the molecule term
        molecules = [i for i in assay_by_molecule if i in superclasses]
        # get the instrument term
        instruments = [i for i in assay_by_sequencer if i in superclasses]
        if len(instruments) == 0:
            instruments = [i for i in assay_by_instrument if i in superclasses]
        # get the measurement for non-molecular readouts
        measurements = [i for i in measurement if i in superclasses]

        readout = {
            "ontology_id": term_id,
            "name": term.name,
            "molecule": _list_to_str(molecules),
            "instrument": _list_to_str(instruments),
            "measurement": _list_to_str(measurements),
        }

        return readout



import os
from pathlib import Path

from rich import print

from bionty.dev._handle_sources import PUBLIC_SOURCES, parse_sources_yaml


def _upload_ontology_artifacts(
    instance: str, lamindb_user: str, lamindb_password: str, python_version: str = "3.9"
) -> None:
    """Uploads all ontology file artifacts to the specified instance.

    Uses all files generated by nox that are found in the _dynamic folder.

    Args:
        instance: The lamindb instance to upload the files to.
        lamindb_user: The lamindb username.
        lamindb_password: The associated password of the lamindb user.
        python_version: The Python version that ran nox. Defaults to '3.9'.
    """
    import lamindb as ln

    _DYNAMIC_PATH = Path(
        f"{os.getcwd()}/.nox/build-package-bionty/lib/python{python_version}/site-packages/bionty/_dynamic"
    )

    ln.setup.login(lamindb_user, password=lamindb_password)
    ln.setup.load(instance, migrate=True)
    with ln.Session() as ss:
        transform = ln.select(
            ln.Transform, name="Bionty ontology artifacts"
        ).one_or_none()

        if transform is None:
            transform = ln.add(ln.Transform(name="Bionty ontology artifacts"))

        run = ln.Run(transform=transform)

        for filename in os.listdir(_DYNAMIC_PATH.absolute()):
            nox_ontology_file_path = os.path.join(_DYNAMIC_PATH.absolute(), filename)

            ontology_ln_file = ss.select(ln.File, name=filename).one_or_none()

            if ontology_ln_file is not None:
                print(
                    "[bold yellow]Found"
                    f" {ontology_ln_file.name}{ontology_ln_file.suffix} on S3."
                    " Skipping ingestion..."
                )
            else:
                ontology_ln_file = ln.File(
                    nox_ontology_file_path, key=filename, run=run
                )
                ss.add(ontology_ln_file)


all_public_versions = parse_sources_yaml(PUBLIC_SOURCES)
# nothing happens if no additional versions are added
if all_public_versions.shape[0] > 23:
    _upload_ontology_artifacts(
        instance="sunnyosun/bionty-assets",
        lamindb_user="testuser2@lamin.ai",
        lamindb_password="goeoNJKE61ygbz1vhaCVynGERaRrlviPBVQsjkhz",
        python_version="3.9",
    )



from pathlib import Path

import pandas as pd

from bionty.dev._handle_sources import LAMINDB_INSTANCE_LOADED
from bionty.dev._io import load_yaml

ROOT_DIR = Path(__file__).parent.resolve()


def display_available_sources() -> pd.DataFrame:
    """Displays all available sources.

    Examples:
        >>> import bionty as bt
        >>> bt.display_available_sources()
    """
    from .dev._handle_sources import LOCAL_SOURCES, parse_sources_yaml

    return parse_sources_yaml(LOCAL_SOURCES).set_index("entity")  # type: ignore


# This function naming is consistent with the `currently_used` field in BiontySource SQL table
# Do not rename!
def display_currently_used_sources() -> pd.DataFrame:
    """Displays all currently used sources.

    Active version is unique for entity + species.

    Examples:
        >>> import bionty as bt
        >>> bt.display_currently_used_sources()
    """
    from .dev._handle_sources import CURRENT_SOURCES, LAMINDB_SOURCES

    VERSIONS_FILE_PATH = (
        LAMINDB_SOURCES if LAMINDB_INSTANCE_LOADED() else CURRENT_SOURCES
    )

    versions = load_yaml(VERSIONS_FILE_PATH.resolve())

    df_rows = []
    for bionty_class, bionty_class_data in versions.items():
        for species, species_data in bionty_class_data.items():
            for source, version in species_data.items():
                df_rows.append(
                    {
                        "entity": bionty_class,
                        "species": species,
                        "source": source,
                        "version": version,
                    }
                )

    return pd.DataFrame(df_rows).set_index("entity")



from typing import Literal, Optional

from .._bionty import Bionty
from ._shared_docstrings import _doc_params, doc_entites


@_doc_params(doc_entities=doc_entites)
class Tissue(Bionty):
    """Tissue.

    1. Uberon
    Edits of terms are coordinated and reviewed on:
    https://github.com/obophenotype/uberon

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        species: str = "all",
        source: Optional[Literal["uberon"]] = None,
        version: Optional[str] = None,
        **kwargs
    ) -> None:
        super().__init__(
            source=source,
            version=version,
            species=species,
            include_id_prefixes={"uberon": ["UBERON"]},
            **kwargs
        )



"""Bionty: Data model generator for biology.

Import the package::

   import bionty as bt

Entities:

.. autosummary::
   :toctree: .

   Species
   Gene
   Protein
   CellMarker
   CellType
   CellLine
   Tissue
   Disease
   Phenotype
   Pathway
   Drug
   Readout
   BFXPipeline

Base model of entity classes:

.. autosummary::
   :toctree: .

   Bionty

Bionty sources:

.. autosummary::
   :toctree: .

    display_available_sources
    display_currently_used_sources
    LOCAL_SOURCES
    reset_sources

External API:

.. autosummary::
   :toctree: .

   Ontology
"""

__version__ = "0.21.0"  # denote release candidate for 0.1.0 with 0.1rc1

# prints warning of python versions
from lamin_logger import py_version_warning

py_version_warning("3.8", "3.10")

from . import _sync_sources

# dynamic classes
from .entities._bfxpipeline import BFXPipeline
from .entities._drug import Drug
from .entities._gene import Gene
from .entities._species import Species
from .entities._protein import Protein
from .entities._cellline import CellLine
from .entities._celltype import CellType
from .entities._cellmarker import CellMarker
from .entities._tissue import Tissue
from .entities._disease import Disease
from .entities._phenotype import Phenotype
from .entities._pathway import Pathway
from .entities._readout import Readout

# tools
from ._bionty import Bionty
from ._ontology import Ontology
from ._display_sources import display_currently_used_sources, display_available_sources

# sources
from .dev._handle_sources import LOCAL_SOURCES, reset_sources

# backward compat
Entity = Bionty



import pandas as pd
import pytest

import bionty as bt


@pytest.fixture(scope="module")
def genes():
    data = {
        "gene symbol": ["A1CF", "A1BG", "FANCD1", "corrupted"],
        "hgnc id": ["HGNC:24086", "HGNC:5", "HGNC:1101", "corrupted"],
        "ensembl_gene_id": [
            "ENSG00000148584",
            "ENSG00000121410",
            "ENSG00000188389",
            "ENSG0000corrupted",
        ],
    }
    df = pd.DataFrame(data).set_index("ensembl_gene_id")

    gn = bt.Gene(source="ensembl", version="release-109")

    return df, gn


def test_inspect_iterable(genes):
    df, gn = genes

    mapping = gn.inspect(df.index, field=gn.ensembl_gene_id)

    expected_mapping = {
        "mapped": ["ENSG00000148584", "ENSG00000121410", "ENSG00000188389"],
        "not_mapped": ["ENSG0000corrupted"],
    }

    assert mapping == expected_mapping


def test_inspect_return_df(genes):
    df, gn = genes

    mapping = gn.inspect(df.index, field=gn.ensembl_gene_id, return_df=True)

    expected_df = pd.DataFrame(
        index=[
            "ENSG00000148584",
            "ENSG00000121410",
            "ENSG00000188389",
            "ENSG0000corrupted",
        ],
        data={
            "__mapped__": [True, True, True, False],
        },
    )

    assert mapping.equals(expected_df)



from __future__ import annotations

import os
from functools import cached_property
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Set, Tuple, Union

import pandas as pd
from lamin_logger import logger
from lamin_logger._lookup import Lookup

from bionty._md5 import verify_md5

from ._ontology import Ontology
from ._settings import check_datasetdir_exists, check_dynamicdir_exists, settings
from .dev._handle_sources import LAMINDB_INSTANCE_LOADED
from .dev._io import s3_bionty_assets, url_download


class Bionty:
    """Bionty base model."""

    def __init__(
        self,
        source: Optional[str] = None,
        version: Optional[str] = None,
        species: Optional[str] = None,
        *,
        include_id_prefixes: Optional[Dict[str, List[str]]] = None,
    ):
        self._fetch_sources()
        # match user input species, source and version with yaml
        self._source_record = self._match_all_sources(
            source=source, version=version, species=species
        )
        self._species = self._source_record["species"]
        self._source = self._source_record["source"]
        self._version = self._source_record["version"]

        # only currently_used sources are allowed inside lamindb instances
        default_sources = list(self._default_sources.itertuples(index=False, name=None))
        if (
            LAMINDB_INSTANCE_LOADED()
            and (self.species, self.source, self.version) not in default_sources
        ):
            logger.error(
                f"Only default sources below are allowed inside LaminDB instances!\n{self._default_sources}\n"  # noqa: E501
            )
            logger.hint(
                "To use a different source, please either:\n    Close your instance"
                " via `lamin close` \n    OR\n    Configure currently_used"
                f" {self.__class__.__name__} source in lnschema_bionty.BiontySource"
            )
            self._source = None  # type: ignore
            return

        self._set_file_paths()
        self.include_id_prefixes = include_id_prefixes

        # df is only read into memory at the init to improve performance
        df = self._load_df()
        # self._df has no index
        if df.index.name is not None:
            df = df.reset_index()
        self._df = df

        # set column names/fields as attributes
        for col_name in self._df.columns:
            try:
                setattr(self, col_name, BiontyField(self, col_name))
            # Some fields of an ontology (e.g. Gene) are not Bionty class attributes and must be skipped.
            except AttributeError:
                pass

    def __repr__(self) -> str:
        # fmt: off
        representation = (
            f"{self.__class__.__name__}\n"
            f"Species: {self.species}\n"
            f"Source: {self.source}, {self.version}\n\n"
            f"📖 {self.__class__.__name__}.df(): ontology reference table\n"
            f"🔎 {self.__class__.__name__}.lookup(): autocompletion of terms\n"
            f"🎯 {self.__class__.__name__}.search(): free text search of terms\n"
            f"🧐 {self.__class__.__name__}.inspect(): check if identifiers are mappable\n"
            f"👽 {self.__class__.__name__}.map_synonyms(): map synonyms to standardized names\n"
            f"🔗 {self.__class__.__name__}.ontology: Pronto.Ontology object"
        )
        # fmt: on
        if self._source is not None:
            return representation
        else:
            return "invalid Bionty object"

    @property
    def species(self):
        """The `name` of `Species` Bionty."""
        return self._species

    @property
    def source(self):
        """Name of the source."""
        return self._source

    @property
    def version(self):
        """The `name` of `version` entity Bionty."""
        return self._version

    @property
    def fields(self) -> Set:
        """All Bionty entity fields."""
        blacklist = {"include_id_prefixes"}
        fields = set(
            [
                field
                for field in vars(self)
                if not callable(getattr(self, field)) and not field.startswith("_")
            ]
        )
        return fields - blacklist

    @cached_property
    def ontology(self):
        """The Pronto Ontology object.

        See: https://pronto.readthedocs.io/en/stable/api/pronto.Ontology.html
        """
        if self._local_ontology_path is None:
            logger.error(f"{self.__class__.__name__} has no Pronto Ontology object!")
            return
        else:
            self._download_ontology_file(
                localpath=self._local_ontology_path,
                url=self._url,
                md5=self._md5,
            )
            return Ontology(handle=self._local_ontology_path)

    def _download_ontology_file(self, localpath: Path, url: str, md5: str = "") -> None:
        """Download ontology file to _local_ontology_path."""
        if not localpath.exists():
            logger.download(f"Downloading {self.__class__.__name__} ontology file...")
            try:
                self._url_download(url, localpath)
            finally:
                # Only verify md5 if it's actually available from the sources.yaml file
                if len(md5) > 0:
                    if not verify_md5(localpath, md5):
                        logger.warning(
                            f"MD5 sum for {localpath} did not match {md5}. Redownloading..."  # noqa: E501
                        )
                        os.remove(localpath)
                        self._url_download(url, localpath)

    def _fetch_sources(self) -> None:
        from ._display_sources import (
            display_available_sources,
            display_currently_used_sources,
        )

        def _subset_to_entity(df: pd.DataFrame, key: str):
            return df.loc[[key]] if isinstance(df.loc[key], pd.Series) else df.loc[key]

        self._default_sources = _subset_to_entity(
            display_currently_used_sources(), self.__class__.__name__
        )

        self._all_sources = _subset_to_entity(
            display_available_sources(), self.__class__.__name__
        )

    def _match_all_sources(
        self,
        source: Optional[str] = None,
        version: Optional[str] = None,
        species: Optional[str] = None,
    ) -> Dict[str, str]:
        """Match a source record base on passed species, source and version."""
        lc = locals()

        # kwargs that are not None
        kwargs = {
            k: lc.get(k)
            for k in ["source", "version", "species"]
            if lc.get(k) is not None
        }
        keys = list(kwargs.keys())

        # if 1 or 2 kwargs are specified, find the best match in all sources
        if (len(kwargs) == 1) or (len(kwargs) == 2):
            cond = self._all_sources[keys[0]] == kwargs.get(keys[0])
            if len(kwargs) == 1:
                row = self._all_sources[cond].head(1)
            else:
                # len(kwargs) == 2
                cond = getattr(cond, "__and__")(
                    self._all_sources[keys[1]] == kwargs.get(keys[1])
                )
                row = self._all_sources[cond].head(1)
        else:
            # if no kwargs are passed, take the currently used source record
            if len(keys) == 0:
                curr = self._default_sources.head(1).to_dict(orient="records")[0]
                kwargs = {
                    k: v
                    for k, v in curr.items()
                    if k in ["species", "source", "version"]
                }
            # if all 3 kwargs are specified, match the record from all sources
            # do the same for the kwargs that obtained from default source to obtain url
            row = self._all_sources[
                (self._all_sources["species"] == kwargs["species"])
                & (self._all_sources["source"] == kwargs["source"])
                & (self._all_sources["version"] == kwargs["version"])
            ].head(1)

        # if no records matched the passed kwargs, raise error
        if row.shape[0] == 0:
            raise ValueError(
                f"No source is available with {kwargs}\nCheck"
                " `bionty.display_available_sources()`"
            )
        return row.to_dict(orient="records")[0]

    @check_dynamicdir_exists
    def _url_download(self, url: str, localpath: Path) -> None:
        """Download file from url to dynamicdir _local_ontology_path."""
        # Try to download from s3://bionty-assets
        s3_bionty_assets(
            filename=self._ontology_filename,
            assets_base_url="s3://bionty-assets",
            localpath=localpath,
        )

        # If the file is not available, download from the url
        if not localpath.exists():
            logger.download(
                f"Downloading {self.__class__.__name__} ontology file from: {url}"
            )
            _ = url_download(url, localpath)

    @check_datasetdir_exists
    def _set_file_paths(self) -> None:
        """Sets version, database and URL attributes for passed database and requested version.

        Args:
            source: The database to find the URL and version for.
            version: The requested version of the database.
        """
        self._url = self._source_record.get("url", "")
        self._md5 = self._source_record.get("md5", "")

        # parquet file name
        self._parquet_filename = f"{self.species}_{self.source}_{self.version}_{self.__class__.__name__}_lookup.parquet"  # noqa: E501
        # parquet file local path
        self._local_parquet_path = settings.dynamicdir / self._parquet_filename
        # ontology file name
        self._ontology_filename = f"{self.species}___{self.source}___{self.version}___{self.__class__.__name__}".replace(
            " ", "_"
        )

        if self._url.endswith(".parquet"):  # user provide reference table as the url
            # no local ontology file
            self._local_ontology_path = None
            if not self._url.startswith("s3://bionty-assets/"):
                self._parquet_filename = None  # type:ignore
        else:
            self._local_ontology_path = settings.dynamicdir / self._ontology_filename

    def _load_df(self) -> pd.DataFrame:
        # Download and sync from s3://bionty-assets
        if self._parquet_filename is None:
            # download url as the parquet file
            self._url_download(self._url, self._local_parquet_path)
        else:
            s3_bionty_assets(
                filename=self._parquet_filename,
                assets_base_url="s3://bionty-assets",
                localpath=self._local_parquet_path,
            )
        # If download is not possible, write a parquet file from ontology
        if not self._local_parquet_path.exists():
            # write df to parquet file
            df = self.ontology.to_df(
                source=self.source, include_id_prefixes=self.include_id_prefixes
            )
            df.to_parquet(self._local_parquet_path)

        # loads the df and reset index
        df = pd.read_parquet(self._local_parquet_path)
        return df

    def df(self) -> pd.DataFrame:
        """Pandas DataFrame of the ontology.

        Returns:
            A Pandas DataFrame of the ontology.

        Examples:
            >>> import bionty as bt
            >>> bt.Gene().df()
        """
        if "ontology_id" in self._df.columns:
            return self._df.set_index("ontology_id")
        else:
            return self._df

    def lookup(self, field: Union[BiontyField, str] = "name") -> Tuple:
        """Return an auto-complete object for the bionty field.

        Args:
            field: The field to lookup the values for.
                   Defaults to 'name'.

        Returns:
            A NamedTuple of lookup information of the field values.

        Examples:
            >>> import bionty as bt
            >>> lookup = bt.Gene().lookup()
            >>> lookup.adgb_dt
            >>> lookup_dict = lookup.dict()
            >>> lookup['ADGB-DT']
        """
        df = self._df

        field = str(field)
        if field not in df.columns:
            raise AssertionError(f"No {field} column exists!")

        return Lookup(
            df=df, field=field, tuple_name=self.__class__.__name__, prefix="bt"
        ).lookup()

    def inspect(
        self, identifiers: Iterable, field: BiontyField, return_df: bool = False
    ) -> Union[pd.DataFrame, Dict[str, List[str]]]:
        """Inspect if a list of identifiers are mappable to the entity reference.

        Args:
            identifiers: Identifiers that will be checked against the Ontology.
            field: The BiontyField of the ontology to compare against.
                          Examples are 'ontology_id' to map against the ontology ID
                          or 'name' to map against the ontologies field names.
            return_df: Whether to return a Pandas DataFrame.

        Returns:
            - A Dictionary that maps the input ontology (keys) to the ontology field (values)
            - If specified A Pandas DataFrame with the curated index and a boolean `__mapped__`
              column that indicates compliance with the default identifier.

        Examples:
            >>> import bionty as bt
            >>> celltype_bionty = bt.CellType()
            >>> celltype_bionty.inspect(["Boettcher cell", "bone marrow cell"], field=celltype_bionty.name)
        """
        mapped_df = pd.DataFrame(index=identifiers)

        try:
            synonyms_mapper = self.map_synonyms(
                identifiers=identifiers, field=field, return_mapper=True
            )
            if len(synonyms_mapper) > 0:
                logger.warning("The identifiers contain synonyms!")
                logger.hint(
                    "To increase mappability, convert them into standardized"
                    " names/symbols using '.map_synonyms()'"
                )
        except Exception:
            pass

        # check if index is compliant
        matches = mapped_df.index.isin(self._df[str(field)])

        # annotated what complies with the default ID
        mapped_df["__mapped__"] = matches

        def unique_rm_empty(idx: pd.Index):
            idx = idx.unique()
            return idx[(idx != "") & (~idx.isnull())]

        mapped = unique_rm_empty(mapped_df.index[mapped_df["__mapped__"]]).tolist()
        unmapped = unique_rm_empty(mapped_df.index[~mapped_df["__mapped__"]]).tolist()

        n_mapped = len(mapped)
        n_unmapped = len(unmapped)
        n_unique_terms = len(mapped) + len(unmapped)
        n_empty = len(matches) - n_unique_terms
        frac_unmapped = round(n_unmapped / len(matches) * 100, 1)
        frac_mapped = 100 - frac_unmapped

        if n_empty > 0:
            logger.warning(
                f"Received {n_unique_terms} unique terms, {n_empty} empty/duplicated"
                " terms are ignored."
            )
        logger.success(f"{n_mapped} terms ({frac_mapped}%) are mapped.")
        logger.warning(f"{n_unmapped} terms ({frac_unmapped}%) are not mapped.")

        if return_df:
            return mapped_df
        else:
            mapping: Dict[str, List[str]] = {}
            mapping["mapped"] = mapped
            mapping["not_mapped"] = unmapped
            return mapping

    def map_synonyms(
        self,
        identifiers: Iterable,
        field: BiontyField,
        *,
        synonyms_field: Union[BiontyField, str] = "synonyms",
        return_mapper: bool = False,
    ) -> Union[Dict[str, str], List[str]]:
        """Maps input identifiers against Ontology synonyms.

        Args:
            identifiers: Identifiers that will be mapped against an Ontology field (BiontyField).
            field: The BiontyField of ontology representing the identifiers.
            return_mapper: Whether to return a dictionary of {identifiers : <mapped field values>}.

        Returns:
            - A list of mapped field values if return_mapper is False.
            - A dictionary of mapped values with mappable identifiers as keys
              and values mapped to field as values if return_mapper is True.

        Examples:
            >>> import bionty as bt
            >>> gene_bionty = bt.Gene(source="ensembl", version="release-108")
            >>> gene_symbols = ["A1CF", "A1BG", "FANCD1", "FANCD20"]
            >>> mapping = gene_bionty.map_synonyms(gene_symbols, gn.symbol)
        """
        from lamin_logger._map_synonyms import map_synonyms

        return map_synonyms(
            df=self._df,
            identifiers=identifiers,
            field=str(field),
            synonyms_field=str(synonyms_field),
            return_mapper=return_mapper,
        )

    def search(
        self,
        string: str,
        field: Union[BiontyField, str] = "name",
        synonyms_field: Union[BiontyField, str, None] = "synonyms",
        case_sensitive: bool = True,
        return_ranked_results: bool = False,
    ) -> pd.DataFrame:
        """Search a given string against a Bionty field.

        Args:
            string: The input string to match against the field ontology values.
            field: The BiontyField of ontology the input string is matching against.
            synonyms_field: Also map against in the synonyms (If None, no mapping against synonyms).
            case_sensitive: Whether the match is case sensitive.
            return_ranked_results: Whether to return all entries ranked by matching ratios.

        Returns:
            Best match record of the input string.

        Examples:
            >>> import bionty as bt
            >>> celltype_bionty = bt.CellType()
            >>> celltype_bionty.search("gamma delta T cell")
        """
        from lamin_logger._search import search

        return search(
            df=self._df,
            string=string,
            field=str(field),
            synonyms_field=str(synonyms_field),
            case_sensitive=case_sensitive,
            return_ranked_results=return_ranked_results,
        )


class BiontyField:
    """Field of a Bionty model."""

    def __init__(self, parent: Bionty, name: str):
        self.parent = parent
        self.name = name

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name



import pandas as pd

import bionty as bt


def test_hp_phenotype_inspect_name():
    df = pd.DataFrame(
        index=[
            "Specific learning disability",
            "Dystonia",
            "Cerebral hemorrhage",
            "Slurred speech",
            "This phenotype does not exist",
        ]
    )

    pt = bt.Phenotype(source="hp", version="2023-01-27")
    inspected_df = pt.inspect(df.index, field=pt.name, return_df=True)

    inspect = inspected_df["__mapped__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)



import shutil

import nox
from laminci import upload_docs_artifact
from laminci.nox import build_docs, run_pre_commit

nox.options.default_venv_backend = "none"


@nox.session
def lint(session: nox.Session) -> None:
    run_pre_commit(session)


@nox.session
@nox.parametrize("group", ["bionty-unit", "bionty-docs"])
def build(session, group):
    session.run(*"pip install .[dev,test]".split())
    coverage_args = "--cov=bionty --cov-append --cov-report=term-missing"  # noqa
    if group == "bionty-unit":
        session.run(*f"pytest -s {coverage_args} ./tests".split())
    elif group == "bionty-docs":
        session.run(*f"pytest -s {coverage_args} ./docs/guide".split())
        shutil.copy("README.md", "docs/README.md")
        build_docs(session)
        upload_docs_artifact(aws=True)



import pandas as pd

import bionty as bt


def test_dron_drug_inspect_name():
    df = pd.DataFrame(
        index=[
            "LILIUM LONGIFLORIUM",
            "citrus bioflavonoids",
            "Ornithine, (L)-Isomer",
            "Hyoscyamus extract",
            "This disease does not exist",
        ]
    )

    dt = bt.Drug(source="dron", version="2023-03-10")
    inspected_df = dt.inspect(df.index, field=dt.name, return_df=True)

    inspect = inspected_df["__mapped__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)



from textwrap import dedent


def _doc_params(**kwds):  # pragma: no cover
    """\
    Docstrings should start with "\" in the first line for proper formatting.
    """

    def dec(obj):
        obj.__orig_doc__ = obj.__doc__
        obj.__doc__ = dedent(obj.__doc__).format_map(kwds)
        return obj

    return dec


def remove_prefix(
    text, prefix
):  # pragma: no cover  # TODO replace with removeprefix Python 3.9+
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


doc_entites = """\
species: `name` of `Species` entity.
        source: The key of the source in the local.yml versions file.
                  Get all available databases with `bionty.display_available_sources`.
        version: The version of the ontology. Typically a date or an actual version.
                  Get available versions with `bionty.display_available_sources`.
"""
species_removed_tmp = "\n".join(doc_entites.split("\n")[1:]).split("\n")
species_removed_tmp[0] = remove_prefix(species_removed_tmp[0], "        ")
species_removed = "\n".join(species_removed_tmp)


doc_curate = """\
df: DataFrame with a column of identifiers
        column: If `column` is `None`, checks the existing index for compliance with
                  the default identifier.
                If `column` denotes an entity identifier, tries to map that identifier
                  to the default identifier.
        field: The type of identifier for mapping.
"""



import pandas as pd
import pytest

import bionty as bt


@pytest.fixture(scope="session")
def genes():
    data = {
        "gene symbol": ["A1CF", "A1BG", "FANCD1", "corrupted"],
        "hgnc id": ["HGNC:24086", "HGNC:5", "HGNC:1101", "corrupted"],
        "ensembl_gene_id": [
            "ENSG00000148584",
            "ENSG00000121410",
            "ENSG00000188389",
            "ENSG0000corrupted",
        ],
    }
    df = pd.DataFrame(data).set_index("ensembl_gene_id")

    gn = bt.Gene(source="ensembl", version="release-109")

    return df, gn


def test_gene_ensembl_inspect_hgnc_id(genes):
    df, gn = genes

    inspected_df = gn.inspect(df["hgnc id"], field=gn.hgnc_id, return_df=True)

    inspect = inspected_df["__mapped__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, False])

    assert inspect.equals(expected_series)



import pandas as pd

import bionty as bt


def test_ensemble_species_inspect_name():
    df = pd.DataFrame(
        index=[
            "spiny chromis",
            "silver-eye",
            "platyfish",
            "california sea lion",
            "This species does not exist",
        ]
    )

    sp = bt.Species(source="ensembl", version="release-109")
    inspected_df = sp.inspect(df.index, field=sp.name, return_df=True)

    inspect = inspected_df["__mapped__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)



import os
import re
import urllib.request
from http.client import BadStatusLine
from pathlib import Path
from urllib.error import HTTPError, URLError

import yaml  # type:ignore

VERSIONS_FILE_PATH = Path(f"{os.getcwd()}/bionty/sources/sources.yaml")


def extract_urls_from_yaml(yaml_file):
    with open(yaml_file, "r") as file:
        yaml_data = yaml.safe_load(file)
        urls = []

        def extract_urls(data):
            if isinstance(data, str):
                urls.extend(
                    re.findall(
                        r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+",
                        data,
                    )
                )
            elif isinstance(data, dict):
                for value in data.values():
                    extract_urls(value)
            elif isinstance(data, list):
                for item in data:
                    extract_urls(item)

        extract_urls(yaml_data)

        return urls


urls = extract_urls_from_yaml(VERSIONS_FILE_PATH)

failed_urls = []
for url in urls:
    try:
        assert urllib.request.urlopen(url, timeout=1000).getcode() == 200
    except (URLError, BadStatusLine):
        print(f"URL: {url} is currently not accessible.")
        pass
    except (AssertionError, ValueError, HTTPError) as e:
        failed_urls.append([url, e])

if len(failed_urls) != 0:
    for fail in failed_urls:
        print(fail)
    raise AssertionError(f"{len(failed_urls)} URLs failed.")
